from questions_three.scaffolds.test_script import \
    precondition, skip, test, test_suite


with test_suite('ExampleSuite'):

    with test('A passing test'):
        assert True, 'That was easy'

    with test('A failing test'):
        assert False, 'Failing is easy too'

    with test('An errant test'):
        raise RuntimeError('Intentional error')

    with test('A skipped test'):
        skip("Let's not run this test.  It's rather silly.")

    with test('A test that skips because its precondition is not met'):
        precondition(False)

    with test('A test that runs because its precondition is met'):
        precondition(True)



from questions_three.exceptions import TestSkipped


def precondition(statement):
    """
    Raise TestSkipped if statement is false
    """
    if not statement:
        raise TestSkipped('Skipping test because precondition is not met')



import os

from twin_sister import dependency

from questions_three.exceptions import InvalidConfiguration

from .transform_type import transform_type


class ModuleCfg:

    def __init__(self, defaults, *, module_name='unidentified module'):
        self._defaults = {k.lower(): v for k, v in defaults.items()}
        self._env = dependency(os).environ
        self._module_name = module_name

    def _from_env(self, name):
        lc_name = name.lower()
        matches = [v for k, v in self._env.items() if lc_name == k.lower()]
        count = len(matches)
        if 0 == count:
            return None
        if 1 == count:
            return matches[0]
        raise InvalidConfiguration(
            'Found more than one environment variable that could '
            'match "%s" (case insensitive)' % name)

    def __getitem__(self, name):
        lc_name = name.lower()
        if lc_name not in self._defaults.keys():
            raise InvalidConfiguration(
                'Refusing to use "%s" from the environment '
                'because it was not defined in a module configuration file '
                'for %s'
                % (name, self._module_name))
        default = self._defaults[lc_name]
        from_env = self._from_env(name)
        return default if from_env is None else transform_type(
            key=name, val=from_env, new_type=type(default))

    def to_dict(self):
        return {
            name: self[name]
            for name in self._defaults.keys()}

    __getattr__ = __getitem__



from datetime import datetime
import weakref

from twin_sister import dependency

from questions_three.logging import logger_for_module
from questions_three.module_cfg import config_for_module
from questions_three.vanilla import format_exception


def call_if_alive(ref, **kwargs):
    func = ref()
    if func:
        func(**kwargs)


def current_time():
    return dependency(datetime).now()


def is_bound(func):
    return hasattr(func, '__self__') and func.__self__


class EventBroker:

    @classmethod
    def get_subscribers(cls):
        """
        Return a set containing all subscriber functions
        """
        found = set()
        for event, refs in cls._subscribers.items():
            found = found | set([r() for r in refs if r()])
        return found

    @classmethod
    def publish(cls, *, event, event_time=None, **kwargs):
        conf = config_for_module(__name__)
        run_id = conf.test_run_id
        if event_time is None:
            event_time = current_time()
        cls._logger.debug(event)
        if event in cls._subscribers.keys():
            for subscriber in cls._subscribers[event]:
                try:
                    call_if_alive(
                        subscriber, event_time=event_time,
                        run_id=run_id, **kwargs)
                except Exception as e:
                    cls._logger.error(format_exception(e))

    @classmethod
    def reset(cls):
        cls._subscribers = {}  # event -> [subscribers]
        cls._logger = logger_for_module(__name__)

    @classmethod
    def subscribe(cls, *, event, func):
        if event not in cls._subscribers.keys():
            cls._subscribers[event] = []
        if is_bound(func):
            subscriber = weakref.WeakMethod(func)
        else:
            subscriber = weakref.ref(func)
        cls._subscribers[event].append(subscriber)


EventBroker.reset()



import sys

from .junit2jenkins_status import junit2jenkins_status

argv = sys.argv
if len(argv) < 2:
    print('Please specify a path which contains JUnit XML reports')
    exit(1)
print(junit2jenkins_status(path=sys.argv[1]))



from datetime import datetime

from twin_sister import dependency


def expand_headers(headers):
    return '\n'.join(['%s: %s' % (k, v) for k, v in headers.items()])


def now():
    return dependency(datetime.utcnow)()


class Request:

    def __init__(self, method, url, headers={}, data=None, **kwargs):
        self.timestamp = now()
        self.method = method.upper()
        self.url = url
        self.headers = headers
        self.data = '\n\n' + str(data) if data else ''

    def __str__(self):
        return (
            ' --------- Request at %s\n' % self.timestamp.isoformat() +
            '%s %s\n' % (self.method, self.url) +
            expand_headers(self.headers) +
            self.data)


class Response:

    def __init__(self, requests_response):
        self.timestamp = now()
        self.response = requests_response

    def __str__(self):
        body = '\n\n%s' % self.response.text if self.response.text else ''
        return (
            ' --------- Response at %s\n' % self.timestamp.isoformat() +
            '\n%d\n' % self.response.status_code +
            expand_headers(self.response.headers) +
            body)


class Transcript:

    def __init__(self):
        self.records = []

    def add_request(self, method,  url, **kwargs):
        self.records.append(Request(method, url, **kwargs))

    def add_response(self, requests_response):
        self.records.append(Response(requests_response))

    def __str__(self):
        return '\n\n'.join([str(r) for r in self.records])



from string import ascii_letters, digits


LETTERS_AND_NUMBERS = ascii_letters + digits


def string_of_sequential_characters(*, character_count):
    if type(character_count) is int and character_count > 0:
        loop_count = 0
        string = ""
        for n in range(character_count):
            if loop_count == len(LETTERS_AND_NUMBERS):
                loop_count = 0
            string += LETTERS_AND_NUMBERS[loop_count]
            loop_count += 1
        return string
    raise TypeError('character_count must be of Type int and greater than 0')



from collections.abc import Mapping
import json
from random import randrange
import sys


def flatten(obj):
    if hasattr(obj, 'to_dict'):
        return obj.to_dict()
    if isinstance(obj, list):
        return [flatten(x) for x in obj]
    return obj


def to_json_serializable(obj):
    if hasattr(obj, 'items'):
        return {k: to_json_serializable(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [to_json_serializable(i) for i in obj]
    if type(obj) in (int, float, complex):
        return obj
    return(str(obj))


class Structure(Mapping):
    """
    Structure of nested key/value pairs, accessible as attributes.
    """

    @classmethod
    def from_json(cls, json_string):
        return cls(**json.loads(json_string))

    @classmethod
    def _inflate_if_mapping(cls, obj):
        if isinstance(obj, list):
            return [cls._inflate_if_mapping(o) for o in obj]
        if {'keys', 'values', 'items'} <= set(dir(obj)):
            return cls(**obj)
        return obj

    def __init__(self, **kwargs):
        self._hash = randrange(sys.maxsize)
        self._members = {
            k: self._inflate_if_mapping(v)
            for k, v in kwargs.items()}

    def __getattr__(self, name):
        if name in self._members.keys():
            return self._members[name]
        raise AttributeError('I have no "%s"' % name)

    def __eq__(self, other):
        if not hasattr(other, 'keys'):
            return False
        my_keys = self.keys()
        other_keys = other.keys()
        return(
            len(my_keys) == len(other_keys) and
            all([k in other_keys for k in my_keys]) and
            all([self[k] == other[k] for k in my_keys])
        )

    def __hash__(self):
        return self._hash

    def __iter__(self):
        return iter(self._members)

    def __getitem__(self, name):
        return self._members[name]

    def __len__(self):
        return len(self._members)

    def __repr__(self):
        return 'Structure: %s' % self.to_dict()

    def __setattr__(self, key, value):
        if key.startswith('_'):
            self.__dict__[key] = value
        else:
            self._members[key] = value

    def __setitem__(self, key, value):
        self._members[key] = value

    def _child_structures(self):
        return [
            val for val in self.values()
            if isinstance(val, self.__class__)]

    def find(self, key):
        if key in self.keys():
            return getattr(self, key)
        for child in self._child_structures():
            val = child.find(key)
            if val:
                return val
        return None

    def items(self):
        return self._members.items()

    def keys(self):
        return self._members.keys()

    def to_dict(self):
        return {
            k: flatten(v)
            for k, v in self.items()}

    def to_json(self):
        try:
            return json.dumps(
                {
                    k: to_json_serializable(v)
                    for k, v in self.to_dict().items()})
        except TypeError as e:
            raise TypeError('%s\n%s' % (e, repr(self))) from e

    def values(self):
        return self._members.values()



from random import randint
from string import ascii_lowercase, digits


BASE36_CHARS = digits + ascii_lowercase


def random_base36_string(length):
    return ''.join([
        BASE36_CHARS[randint(0, len(BASE36_CHARS)-1)]
        for n in range(length)])



from datetime import datetime
from subprocess import Popen, PIPE, STDOUT
import sys

from twin_sister import dependency


# Flush process output streams every this number of seconds
FLUSH_INTERVAL = 3 * 60


def now():
    return dependency(datetime).now()


class Job:
    def __init__(self, *, path_to_script, output_stream):
        self._last_flush = None
        self._output_stream = output_stream
        self._path_to_script = path_to_script
        self._proc = None
        self.name = path_to_script

    def _flush(self):
        output = self._proc.stdout.read().decode('utf-8')
        if output.strip():
            self._output_stream.write(
                '\n** From %s **\n' % self._path_to_script)
            self._output_stream.write('%s\n' % output)
            self._output_stream.flush()
        self._last_flush = now()

    def _should_flush(self):
        return (now() - self._last_flush).total_seconds() > FLUSH_INTERVAL

    def run(self):
        self._proc = dependency(Popen)(
            [dependency(sys.executable), self._path_to_script],
            stderr=STDOUT, stdout=PIPE)
        self._last_flush = now()

    def kill(self):
        self._proc.kill()

    def poll(self):
        """
        Flush output if necessary.
        Periodic flushing prevents us from deadlocking if an output
          buffer becomes full.
        Return process exit code (or None if still running)
        """
        code = self._proc.poll()
        if code is not None or self._should_flush():
            self._flush()
        return code



from expects import expect, contain, equal
from questions_three.http_client import HttpClient
from questions_three.scaffolds.check_script import check, check_suite


with check_suite('SimpleHttpCheck'):

    response = HttpClient().get('http://example.com/')

    with check('Verify status code is 200'):
        expect(response.status_code).to(equal(200))

    with check('Verify response contains expected text'):
        expect(response.text).to(contain('illustrative examples'))



from contextlib import contextmanager

from twin_sister import dependency

from questions_three.constants import TestEvent
from questions_three.event_broker import EventBroker
from questions_three.exceptions import TestSkipped


@contextmanager
def test(name):
    broker = dependency(EventBroker)
    broker.publish(event=TestEvent.test_started, test_name=name)
    try:
        yield
    except TestSkipped as e:
        broker.publish(
            event=TestEvent.test_skipped, test_name=name, exception=e)
    except AssertionError as e:
        broker.publish(
            event=TestEvent.test_failed, test_name=name, exception=e)
    except Exception as e:
        broker.publish(
            event=TestEvent.test_erred, test_name=name, exception=e)
    broker.publish(event=TestEvent.test_ended, test_name=name)



import json

from .structure import Structure


def structures_from_json(json_string):
    return [Structure(**dct) for dct in json.loads(json_string)]


def structures_to_json(sequence_of_structures):
    return '[' + ','.join([s.to_json() for s in sequence_of_structures]) + ']'



"""
Due to some hocus-pocus which you can learn from the source code,
simply declaring the class causes the test to run, so this file
is a plain-old python executable.
"""
from questions_three.scaffolds.xunit import TestSuite, skip


class MyXunitSuite(TestSuite):

    def setup_suite(self):
        """
        Perform setup that affects all tests here
        Changes to "self" will affect all tests.
        """
        print('This runs once at the start of the suite')

    def teardown_suite(self):
        print('This runs once at the end of the suite')

    def setup(self):
        """
        Perform setup for each test here.
        Changes to "self" will affect the current test only.
        """
        print('This runs before each test')

    def teardown(self):
        print('This runs after each test')

    def test_that_passes(self):
        """
        The method name is xUnit magic.
        Methods named "test..." get treated as test cases.
        """
        print('This test passes')

    def test_that_fails(self):
        print('This test fails')
        assert False, 'I failed'

    def test_that_errs(self):
        print('This test errs')
        raise RuntimeError('I tried to think but nothing happened')

    def test_that_skips(self):
        print('This test skips')
        skip("Don't do that")



import logging
from logging import Formatter, INFO, Logger
import os

from twin_sister import dependency

from .constants import MESSAGE_FORMAT


def _parent_module_name(module_name):
    parts = module_name.split('.')
    if len(parts) > 1:
        return '.'.join(parts[:-1])
    return None


def _level_for_module(module_name, environ):
    var_name = '%s_LOG_LEVEL' % module_name.replace('.', '_').upper()
    if var_name in environ.keys():
        level = getattr(logging, environ[var_name])
    else:
        parent_module = _parent_module_name(module_name)
        if parent_module:
            level = _level_for_module(parent_module, environ=environ)
        else:
            level = INFO
    return level


def logger_for_module(module_name):
    """
    Return a Logger for the module with the given name.
    Its level can be controlled by an environment variable:
      <module_name.upper()>_LOG_LEVEL=<DEBUG | INFO | WARNING | ERROR>

    module_name -- (str)  Name of the module
    """
    environ = dependency(os).environ
    handler = dependency(logging).StreamHandler()
    handler.setFormatter(Formatter(fmt=MESSAGE_FORMAT))
    logger = Logger(name=module_name)
    logger.addHandler(handler)
    logger.setLevel(_level_for_module(module_name, environ=environ))
    return logger



import os


def module_filename(module, filename):
    """
    Return the full path to a specific file inside a module
    """
    path, _ = os.path.split(module.__file__)
    return os.path.join(path, filename)



import os
import traceback


def path_to_entry_script():
    """
    Return the full path and filename of the script that was called from
    the command line, or None if the interpreter was launched without a
    script.
    """
    stack = traceback.extract_stack()
    if stack:
        fn = stack[0].filename
        if not fn.startswith('<'):
            return os.path.join(os.getcwd(), fn)
    return None



from enum import Enum


TestEvent = Enum(
    "TestEvent",
    """
    artifact_created
    report_created
    suite_ended
    suite_erred
    suite_results_compiled
    suite_started
    test_ended
    test_failed
    test_erred
    test_skipped
    test_started
    """)



def join(one, two):
    if not one:
        return two
    if one.endswith('/') and two.startswith('/'):
        return one + two[1:]
    if one.endswith('/') or two.startswith('/'):
        return one + two
    return '%s/%s' % (one, two)


def url_append(*parts):
    """
    Replacement for urljoin which
     1. Does not eliminate characters when slashes are present
     2. Joins an arbitrary number of parts
    """
    whole = ''
    for part in [p for p in parts if p]:
        whole = join(whole, part)
    return whole



from questions_three.constants import TestEvent
from questions_three.event_broker import EventBroker
from questions_three.exceptions import TestSkipped

from .runnable_suite import RunnableSuite


class TestRunner:
    def __init__(self, suite_name, suite_attributes, test_name):
        self.suite_name = suite_name
        self.suite_attributes = suite_attributes
        self.test_name = test_name

    def publish(self, event, **kwargs):
        EventBroker.publish(
            event=event, suite_name=self.suite_name,
            test_name=self.test_name, **kwargs)

    def run(self):
        self.publish(TestEvent.test_started)
        clean_suite = RunnableSuite(attrs=self.suite_attributes)
        try:
            clean_suite.run_one_test(test_name=self.test_name)
        except TestSkipped as e:
            self.publish(TestEvent.test_skipped, exception=e)
        except AssertionError as e:
            self.publish(TestEvent.test_failed, exception=e)
        except Exception as e:
            self.publish(TestEvent.test_erred, exception=e)
        self.publish(TestEvent.test_ended)



import os
from xml.etree import ElementTree

from twin_sister import dependency

from .status import ERROR, FAIL, PASS


def locate_suite(root):
    return root if 'testsuite' == root.tag \
        else root.find(".//testsuite")


def extract_status(filename):
    with dependency(open)(filename, 'r') as f:
        xml = f.read().strip()
    root = ElementTree.fromstring(xml)
    suite = locate_suite(root)
    if suite is not None:
        keys = suite.attrib.keys()
        if 'failures' in keys and int(suite.attrib['failures']):
            return FAIL
        if 'errors' in keys and int(suite.attrib['errors']):
            return ERROR
    return PASS


def walk(path):
    return dependency(os.walk)(path)


def junit2jenkins_status(*, path):
    erred = False
    for dirpath, dirnames, filenames in walk(path):
        for fn in filenames:
            if not fn.lower().endswith('.xml'):
                continue
            status = extract_status(os.path.join(dirpath, fn))
            if ERROR == status:
                # We still need to look for failures
                erred = True
            if FAIL == status:
                # One failure fails the whole lot
                return FAIL
    return ERROR if erred else PASS



from questions_three.scaffolds.common import skip

from .test import test


def funcname(f):
    return f.__name__.replace('_', ' ')


def identify(thing):
    attrs = dir(thing)
    if 'func' in attrs and 'args' in attrs:
        return '%s with %s' % (funcname(thing.func), thing.args)
    return funcname(thing)


def do_and_check(*, do, checks):
    """
    Perform one action, followed by a series of checks, each as its own test
    do (function): Execute this
    checks (sequence of functions): Execute each of these as its own test
    """
    err_from_do = None
    try:
        do()
    except Exception as e:
        err_from_do = e
    for check in checks:
        with test('%s and %s' % (identify(do), identify(check))):
            if err_from_do:
                skip('Skipping check because "do" function failed')
            else:
                check()
    if err_from_do:
        raise err_from_do



from questions_three.constants import TestEvent
from questions_three.exceptions import TestSkipped
from questions_three.event_broker import EventBroker
from questions_three.scaffolds.common.activate_reporters \
    import activate_reporters
from questions_three.scaffolds.common.configure_logging \
    import configure_logging

from .runnable_suite import RunnableSuite
from .test_runner import TestRunner


class SuiteRunner:

    def __init__(self, suite_name, suite_attributes):
        self.suite_name = suite_name
        self.suite_attributes = suite_attributes
        configure_logging()
        activate_reporters()

    def discover_tests(self):
        return [
            k for k, v in self.suite_attributes.items()
            if k.startswith('test')]

    def publish(self, event, **kwargs):
        EventBroker.publish(event=event, suite_name=self.suite_name, **kwargs)

    def run(self):
        self.publish(TestEvent.suite_started)
        stateful_suite = RunnableSuite(self.suite_attributes)
        suite_erred = False
        suite_skipped_exception = None
        try:
            stateful_suite.setup_suite()
        except TestSkipped as e:
            suite_skipped_exception = e
        except Exception as e:
            self.publish(TestEvent.suite_erred, exception=e)
            suite_erred = True
        for test_name in self.discover_tests():
            if suite_erred:
                self.publish(
                    TestEvent.test_skipped, test_name=test_name,
                    exception=TestSkipped('Suite setup failed'))
            elif suite_skipped_exception:
                self.publish(
                    TestEvent.test_skipped, test_name=test_name,
                    exception=suite_skipped_exception)
            else:
                TestRunner(
                    suite_name=self.suite_name,
                    suite_attributes=dict(
                        self.suite_attributes,
                        **stateful_suite.get_suite_context()),
                    test_name=test_name).run()
        try:
            if not suite_skipped_exception:
                stateful_suite.teardown_suite()
        except Exception as e:
            self.publish(TestEvent.suite_erred, exception=e)
        self.publish(TestEvent.suite_ended)



import boto3
from twin_sister import dependency

from questions_three.exceptions import InvalidConfiguration
from questions_three.logging import logger_for_module
from questions_three.module_cfg import config_for_module
from questions_three.reporters.artifact_saver import ArtifactSaver
from questions_three.reporters.artifact_saver.artifact_saver import \
    sanitize_filename


class S3ArtifactSaver(ArtifactSaver):

    def __init__(self):
        super().__init__()
        self._suite_name = None

    def save(self, *, artifact, path, filename, content_type):
        log = logger_for_module(__name__)
        bucket = config_for_module(__name__).s3_bucket_for_artifacts
        if not bucket:
            raise InvalidConfiguration('$S3_BUCKET_FOR_ARTIFACTS is not set')
        client = dependency(boto3).client('s3')
        kwargs = {
            'Bucket': bucket,
            'Key': f'{path}/{filename}',
            'Body': artifact}
        if content_type:
            kwargs['ContentType'] = content_type
        log.debug(f'PUT {bucket}:{path}/{filename}')
        client.put_object(**kwargs)

    def path_for_artifact(self, suite_name, test_name, run_id):
        if suite_name is None:
            suite_name = self._suite_name or '_no_suite_supplied'
        if test_name is None:
            test_name = '_no_test_supplied'
        return '%s/artifacts/%s/%s' % (
            run_id,
            sanitize_filename(suite_name),
            sanitize_filename(test_name))

    def on_artifact_created(
            self, run_id, artifact, artifact_group=None, artifact_type=None,
            artifact_mime_type=None, suite_name=None,
            test_name=None, **kwargs):
        log = logger_for_module(__name__)
        path = self.path_for_artifact(suite_name, test_name, run_id)
        filename = self.filename_for_artifact(
            artifact_group=artifact_group,
            artifact_type=artifact_type,
            mime_type=artifact_mime_type)
        log.debug(f'Saving artifact {filename}')
        self.save(
            artifact=artifact, path=path, filename=filename,
            content_type=artifact_mime_type)

    def on_report_created(
            self, report_filename, report_content, run_id, **kwargs):
        log = logger_for_module(__name__)
        log.debug(f'Saving report {report_filename}')
        self.save(
            artifact=report_content,
            path=run_id,
            filename=report_filename,
            content_type=None)



from datetime import datetime, timedelta
import time

from twin_sister import dependency


def now():
    return dependency(datetime).now()


def wait_for(func, *, timeout, throttle):
    """
        Execute a function and retry if the return value is not truthy
        func -- (callable) Execute this.
        timeout -- (number) Raise TimeoutError after this number of seconds
        throttle -- (number) Wait this number of seconds between retries

        Return whatever func returns
        """
    sleep = dependency(time.sleep)
    expiry = now() + timedelta(seconds=timeout)
    while True:
        if func():
            return
        if now() > expiry:
            raise TimeoutError(f'Exceeded wait time of {timeout} seconds')
        sleep(throttle)



import logging

from twin_sister import dependency

from questions_three.logging.constants import \
    MESSAGE_FORMAT as LOG_MESSAGE_FORMAT


def configure_logging():
    dependency(logging).basicConfig(
        format=LOG_MESSAGE_FORMAT,
        level=logging.INFO)



from .call_with_exception_tolerance import call_with_exception_tolerance  # noqa
from .func_that_raises import func_that_raises  # noqa: F401
from .format_exception import format_exception  # noqa: F401
from .module_filename import module_filename  # noqa: F401
from .mutable_object import MutableObject  # noqa: F401
from .path_to_entry_script import path_to_entry_script  # noqa: F401
from .random_base36_string import random_base36_string  # noqa: F401
from .string_of_sequential_characters import string_of_sequential_characters  # noqa:
from .structure import Structure  # noqa: F401
from .structure_sequence import structures_from_json, structures_to_json  # noqa
from .url_append import url_append  # noqa: F401
from .wait_for import wait_for  # noqa: F401



from datetime import datetime, timedelta
import time

from twin_sister import dependency


def now():
    return dependency(datetime).now()


def call_with_exception_tolerance(func, *, tolerate, timeout, throttle):
    """
    Execute a function and retry if a given exception is raised

    func -- (callable) Execute this.
    tolerate -- (Exception or sequence of Exception) Retry if this is raised
    timeout -- (number) Re-raise the exception after this number of seconds
    throttle -- (number) Wait this number of seconds between retries

    Return whatever func returns
    """
    # Convert all sequences to tuples so we can use them in an except statement
    if hasattr(tolerate, '__iter__'):
        tolerate = tuple(tolerate)
    sleep = dependency(time.sleep)
    expiry = now() + timedelta(seconds=timeout)
    while True:
        try:
            return func()
        except tolerate:
            if now() > expiry:
                raise
        sleep(throttle)



import re

from questions_three.constants import TestEvent
from questions_three.exceptions import UndefinedEvent

from .event_broker import EventBroker

EVENT_HANDLER_PATTERN = re.compile('^on_(.+)$')


def event_for_handler(handler_name):
    mat = EVENT_HANDLER_PATTERN.search(handler_name)
    if not mat:
        raise RuntimeError('Failed to parse %s' % handler_name)
    event = mat.group(1)
    if event in [e.name for e in TestEvent]:
        return TestEvent[mat.group(1)]
    raise UndefinedEvent('"%s" is not defined as a test event' % event)


def event_handler_names(obj):
    return [name for name in dir(obj) if name.startswith('on_')]


def subscribe_event_handlers(obj):
    """
    Detect event handlers in an object and subscribe each to its event.
    """
    for func_name in event_handler_names(obj):
        attr = getattr(obj, func_name)
        if hasattr(attr, '__call__'):
            EventBroker.subscribe(
                event=event_for_handler(func_name),
                func=getattr(obj, func_name))



class Status:

    def __init__(self, level, name):
        self.level = level
        self.name = name

    def __eq__(self, other):
        return hasattr(other, 'level') and self.level == other.level

    def __ge__(self, other):
        return self.level >= other.level

    def __gt__(self, other):
        return self.level > other.level

    def __le__(self, other):
        return self.level <= other.level

    def __lt__(self, other):
        return self.level < other.level

    def __str__(self):
        return self.name

    def __repr__(self):
        return('<Jenkins status %s>' % self.name)


PASS = Status(0, 'SUCCESS')
ERROR = Status(1, 'UNSTABLE')
FAIL = Status(2, 'FAILURE')



import os
import sys

from .get_search_path import get_search_path
from .run_all import run_all

search_path = get_search_path()
if not os.path.exists(search_path):
    sys.stderr.write('Failed to locate "%s"\n' % search_path)
    exit(1)
exit(run_all(search_path))



from contextlib import contextmanager

from twin_sister import dependency

from questions_three.constants import TestEvent
from questions_three.event_broker import EventBroker
from questions_three.scaffolds.common.configure_logging \
    import configure_logging


@contextmanager
def test_suite(name):
    broker = dependency(EventBroker)
    configure_logging()
    broker.publish(event=TestEvent.suite_started, suite_name=name)
    try:
        yield
    except Exception as e:
        broker.publish(
            event=TestEvent.suite_erred, exception=e, suite_name=name)
    broker.publish(event=TestEvent.suite_ended, suite_name=name)



import mimetypes
import os
from string import digits, ascii_letters

from twin_sister import dependency

from questions_three.event_broker.handler_detection \
    import subscribe_event_handlers
from questions_three.module_cfg import config_for_module
from questions_three.vanilla import random_base36_string


def extension_for_type(mime_type):
    """
    mimetypes.guess_extension introduces some bizarre randomness.
    This function removes it for the extensions that interest us most.
    """
    ext = None
    if 'image/png' == mime_type:
        ext = '.png'
    elif 'text/plain' == mime_type:
        ext = '.txt'
    elif 'text/html' == mime_type:
        ext = '.html'
    elif 'application/json' == mime_type:
        ext = '.json'
    elif mime_type:
        ext = mimetypes.guess_extension(mime_type)
    return ext or '.bin'


def sanitize_filename(ugly):
    return ''.join(map(
        lambda c: c if c in ascii_letters + digits + '.-_' else '_',
        ugly))


class ArtifactSaver:

    def __init__(self):
        self._suite_name = 'suiteless'

    def activate(self):
        subscribe_event_handlers(self)

    def reports_path(self):
        config = config_for_module(__name__)
        return config.reports_path

    def save(self, *, artifact, path, filename, mode):
        full_path = os.path.join(self.reports_path(), path)
        dependency(os).makedirs(full_path, exist_ok=True)
        with dependency(open)(
                os.path.join(full_path, filename), mode) as f:
            f.write(artifact)

    def path_for_artifact(self, suite_name, test_name):
        if suite_name is None:
            suite_name = self._suite_name if test_name else ''
        if test_name is None:
            test_name = ''
        return os.path.join(
            'artifacts',
            sanitize_filename(suite_name),
            sanitize_filename(test_name))

    @staticmethod
    def filename_for_artifact(
            *, artifact_group, artifact_type, mime_type):
        if artifact_type is None:
            artifact_type = ''
        return '%s%s-%s%s' % (
            artifact_group + '-' if artifact_group else '',
            artifact_type or 'artifact',
            random_base36_string(length=6),  # collision: 1 in 2.2 billion
            extension_for_type(mime_type))

    def on_artifact_created(
            self, artifact, artifact_group=None, artifact_type=None,
            artifact_mime_type=None, suite_name=None,
            test_name=None, **kwargs):
        path = self.path_for_artifact(suite_name, test_name)
        filename = self.filename_for_artifact(
            artifact_group=artifact_group,
            artifact_type=artifact_type,
            mime_type=artifact_mime_type)
        mode = 'wb' if 'decode' in dir(artifact) else 'w'
        self.save(artifact=artifact, path=path, filename=filename, mode=mode)

    def on_report_created(self, report_filename, report_content, **kwargs):
        self.save(
            artifact=report_content,
            path='',
            filename=sanitize_filename(report_filename),
            mode='w')

    def on_suite_started(self, suite_name, **kwargs):
        self._suite_name = suite_name



from questions_three.exceptions import InvalidHttpResponse


def extract_location_header(response):
    try:
        location = response.headers['Location']
    except KeyError as e:
        raise InvalidHttpResponse(
            f'HTTP {response.status_code} response '
            'lacks a location header') from e
    if '' == location:
        raise InvalidHttpResponse(
            f'HTTP {response.status_code} response '
            'contains an empty location header')
    return location



from .suite_runner import SuiteRunner


class MetaSuite(type):

    def __new__(cls, name, parents, dct):
        # Don't treat the base class like a suite
        if 'TestSuite' != name:
            SuiteRunner(suite_name=name, suite_attributes=dct).run()
        return type.__new__(cls, name, parents, dct)



from datetime import datetime
import sys
from time import sleep

from questions_three.logging import logger_for_module
from questions_three.module_cfg import config_for_module
from twin_sister import dependency

from .job import Job


class Pool:

    def __init__(self, limit):
        self._limit = limit
        self._running = []
        self._log = dependency(logger_for_module)(__name__)
        self._stdout = dependency(sys.stdout)
        self.failure_count = 0

    def _prune(self):
        running = []
        for job in self._running:
            exit_code = job.poll()
            if exit_code is None:
                running.append(job)
            elif exit_code > 0:
                self.failure_count += 1
        self._running = running
        self._log.debug('In progress: %s' % str([job.name for job in running]))

    def add(self, suite_filename):
        self._log.info('Executing %s\n' % suite_filename)
        job = Job(path_to_script=suite_filename, output_stream=self._stdout)
        job.run()
        self._running.append(job)

    def _kill_jobs(self):
        for job in self._running:
            job.kill()

    def wait_for_jobs(self, *, expiry):
        cfg = config_for_module(__name__)
        throttle = cfg.delay_between_checks_for_parallel_suite_completion
        while self._running:
            if expiry and dependency(datetime).now() > expiry:
                self._kill_jobs()
                raise TimeoutError('Tests ran longer than the configured limit')
            sleep(throttle)
            self._prune()

    def has_capacity(self):
        self._prune()
        return self._limit is None or len(self._running) < self._limit



import requests
from twin_sister import dependency

from questions_three.constants import TestEvent
from questions_three.event_broker import EventBroker, subscribe_event_handlers
from questions_three.exceptions import TooManyRedirects
from questions_three.logging import logger_for_module
from questions_three.module_cfg import config_for_module

from .construct_redirect_url import construct_redirect_url
from .extract_location_header import extract_location_header
from .inspect_response import inspect_response
from .transcript import Transcript

HTTP_METHODS = ('delete', 'get', 'head', 'options', 'patch', 'post', 'put')

# See https://tools.ietf.org/html/rfc7231#section-6.4.4
# and https://tools.ietf.org/html/rfc7238 for 308
HANDLE_THESE_REDIRECT_STATUS_CODES = (301, 302, 303, 307, 308)

STOP_FOLLOWING_REDIRECTS_AFTER_THIS_MANY = 5


class HttpClient:

    def __init__(self):
        subscribe_event_handlers(self)
        config = config_for_module(__name__)
        self._proxies = {}
        if config.http_proxy:
            self._proxies['http'] = config.http_proxy
        if config.https_proxy:
            self._proxies['https'] = config.https_proxy
        self._logger = logger_for_module(__name__)
        self._session = None
        self._persistent_headers = {}
        self._transcript = Transcript()
        self._verify_certs = config.https_verify_certs

    def enable_cookies(self):
        self._session = dependency(requests).Session()

    def set_persistent_headers(self, **kwargs):
        """
        Instruct the client to send the specified headers with each request

        Specify headers as keyword arguments (name1=val1, name2=val2...)
        """
        for k, v in kwargs.items():
            if v is None:
                if k in self._persistent_headers.keys():
                    del self._persistent_headers[k]
            else:
                self._persistent_headers[k] = v

    def on_suite_erred(self, suite_name=None, test_name=None, **kwargs):
        EventBroker.publish(
            event=TestEvent.artifact_created,
            artifact=str(self._transcript),
            artifact_mime_type='text/plain',
            artifact_type='http_transcript',
            suite_name=suite_name,
            test_name=test_name)

    on_test_erred = on_suite_erred
    on_test_failed = on_suite_erred

    def _send_plain_request(self, method, *args, **kwargs):
        kwargs['allow_redirects'] = False
        if self._proxies and 'proxies' not in kwargs.keys():
            kwargs['proxies'] = self._proxies
        func = getattr(dependency(requests), method)
        return func(*args, timeout=self._socket_timeout(), **kwargs)

    def _send_session_request(
            self, method, url, *, verify, proxies=None, **kwargs):
        request = requests.Request(method.upper(), url, **kwargs)
        prepped = self._session.prepare_request(request)
        return self._session.send(
            prepped, allow_redirects=False, verify=verify,
            proxies=proxies or self._proxies,
            timeout=self._socket_timeout())

    @staticmethod
    def _check_request_kwargs(kwargs):
        if 'json' in kwargs.keys():
            raise NotImplementedError(
                'I do not support "json" because I am a transport component. '
                'Consider specifying "data=json.dumps(thing)" instead.')

    def _socket_timeout(self):
        config = config_for_module(__name__)
        timeout = config.http_client_socket_timeout
        if timeout == '':
            timeout = None
        if timeout is not None:
            try:
                timeout = float(timeout)
            except ValueError as e:
                raise TypeError(
                    'expected $CLIENT_SOCKET_TIMEOUT '
                    f'("{timeout}") to be a number'
                    ) from e
        return timeout

    def _request(
            self, method, url, headers={},
            redirect_depth=0, **kwargs):
        self._check_request_kwargs(kwargs)
        headers = dict(self._persistent_headers, **headers)
        self._transcript.add_request(
            method, url, headers=headers, **kwargs)
        self._logger.debug('%s %s' % (method.upper(), url))
        if self._session is None:
            func = self._send_plain_request
        else:
            func = self._send_session_request
        resp = func(
            method, url, headers=headers, verify=self._verify_certs,
            **kwargs)
        self._logger.debug('HTTP %d\n%s' % (resp.status_code, resp.text))
        self._transcript.add_response(resp)
        if resp.status_code in HANDLE_THESE_REDIRECT_STATUS_CODES:
            if redirect_depth >= STOP_FOLLOWING_REDIRECTS_AFTER_THIS_MANY - 1:
                raise TooManyRedirects(
                    'Detected likely infinite HTTP redirection')
            return self._request(
                method,
                construct_redirect_url(
                    request_url=url,
                    response_location_header=extract_location_header(resp)),
                headers=headers, redirect_depth=redirect_depth+1, **kwargs)
        inspect_response(resp)
        return resp

    def _request_func(self, method):
        def f(*args, **kwargs):
            return self._request(method, *args, **kwargs)
        return f

    def __getattr__(self, name):
        if name in HTTP_METHODS:
            return self._request_func(name)
        raise AttributeError('I have no "%s"' % name)



import questions_three.exceptions.http_error as error


def extract_error_classes():
    return [
        getattr(error, name)
        for name in dir(error)
        if not name.startswith('_')]


ERROR_CLASSES = [
    cls for cls in extract_error_classes()
    if issubclass(cls, error.HttpError)]


STATUS_GROUP_HEADS = {
    cls.status_group: cls
    for cls in ERROR_CLASSES
    if cls.status_group is not None and cls.status_code is None}


BY_STATUS_CODE = {
    cls.status_code: cls
    for cls in ERROR_CLASSES
    if cls.status_code}


def class_matching_status_code(code):
    if code in BY_STATUS_CODE.keys():
        return BY_STATUS_CODE[code]
    return None


def group_head_matching_status_code(code):
    group = int(code / 100)
    if group in STATUS_GROUP_HEADS.keys():
        return STATUS_GROUP_HEADS[group]
    return None


def inspect_response(resp):
    code = int(resp.status_code)
    cls = (
        class_matching_status_code(code) or
        group_head_matching_status_code(code))
    if cls:
        raise cls(resp.text, response=resp)



import importlib
import re

from twin_sister import dependency

from questions_three.exceptions import InvalidConfiguration
from questions_three.module_cfg import config_for_module
from questions_three.reporters.artifact_saver import ArtifactSaver
from questions_three.reporters.event_logger import EventLogger
from questions_three.reporters.junit_reporter import JunitReporter
from questions_three.reporters.result_compiler import ResultCompiler
from questions_three.reporters.s3_artifact_saver import S3ArtifactSaver


BUILT_IN_REPORTERS = (
    ArtifactSaver, EventLogger, JunitReporter, ResultCompiler, S3ArtifactSaver)

_active_reporters = []

enabled = True


def custom_reporter_filename():
    conf = config_for_module(__name__)
    return conf.custom_reporters_file


def custom_reporter_class_names():
    filename = custom_reporter_filename()
    if filename:
        with dependency(open)(filename, 'r') as f:
            return [
                line
                for line in [l.strip() for l in f.read().split('\n')]
                if line and not line.startswith('#')]
    return []


MODULE_AND_CLASS_PATTERN = re.compile(r'^(.*?)\.([^\.]+)$')


def import_class(full_name):
    mat = MODULE_AND_CLASS_PATTERN.match(full_name)
    if not mat:
        raise InvalidConfiguration(
            'Failed to parse line in %s: "%s"'
            % (custom_reporter_filename(), full_name))
    module_name, class_name = mat.groups()
    module = dependency(importlib).import_module(module_name)
    return getattr(module, class_name)


def configured_reporters():
    conf = config_for_module(__name__)
    configured_names = conf.event_reporters.split(',')
    return [
        cls for cls in BUILT_IN_REPORTERS
        if cls.__name__ in configured_names
    ]


def custom_reporters():
    return [import_class(name) for name in custom_reporter_class_names()]


def activate_reporters():
    if not enabled:
        return
    for cls in configured_reporters() + custom_reporters():
        reporter = cls()
        reporter.activate()
        # Create a strong reference so the subscriber stays alive
        _active_reporters.append(reporter)



import os
import sys

from questions_three.exceptions import NoSuchRecord
from questions_three.logging import logger_for_module
from questions_three.vanilla import module_filename
import yaml
from twin_sister import dependency

from .module_cfg import ModuleCfg


# If True, use injected "open" function.
# This hack eliminates the need for unrelated unit tests to manage fake
#  module config files.
UNIT_TEST_MODE = False
VALID_TYPES = (bool, float, int, str, type(None))
YML_FILENAME = 'module_cfg.yml'


def validate_defaults(defaults, *, filename):
    for k, v in defaults.items():
        if type(v) not in VALID_TYPES:
            raise TypeError(
                '%s: %s is not in the valid types list: %s'
                % (filename, k, VALID_TYPES))


def contents_of_file(filename):
    if UNIT_TEST_MODE:
        exists = dependency(os.path).exists
        fopen = dependency(open)
    else:
        exists = os.path.exists
        fopen = open
    if exists(filename):
        with fopen(filename, 'r') as f:
            return f.read()
    return ''


def parent_module(module):
    parts = module.__name__.split('.')
    if len(parts) > 1:
        name = '.'.join(parts[:-1])
        return dependency(sys).modules[name]
    return None


def inherit_missing_defaults(*, defaults, parent_module):
    return dict(defaults_for_module(parent_module), **defaults)


def defaults_for_module(module):
    log = logger_for_module(__name__)
    filename = module_filename(module, YML_FILENAME)
    defaults = yaml.load(contents_of_file(filename), Loader=yaml.Loader) or {}
    log.debug('Configuration from %s: %s' % (filename, defaults))
    validate_defaults(defaults, filename=filename)
    parent = parent_module(module)
    if parent:
        defaults = inherit_missing_defaults(
            defaults=defaults, parent_module=parent)
    return defaults


def config_for_module(module_name):
    modules = dependency(sys).modules
    if module_name not in modules.keys():
        raise NoSuchRecord('Found no module called "%s"' % module_name)
    return ModuleCfg(
        defaults=defaults_for_module(modules[module_name]),
        module_name=module_name)



from datetime import datetime
from junit_xml import TestCase, TestSuite
import os
from twin_sister import dependency
from xml.etree import ElementTree

from questions_three.constants import TestEvent, TestStatus
from questions_three.event_broker import EventBroker, subscribe_event_handlers
from questions_three.module_cfg import config_for_module
from questions_three.vanilla import format_exception,  path_to_entry_script


def convert_status(status):
    if TestStatus.erred == status:
        return 'error'
    if TestStatus.failed == status:
        return 'failure'
    if status in (None, TestStatus.passed):
        return None
    return status.name


def current_time():
    return dependency(datetime).now()


def exception_str(e):
    if e:
        return '(%s) %s' % (type(e), e)
    return ''


def extract_timestamp(test_result):
    if test_result.start_time:
        return test_result.start_time.isoformat()
    return None


def convert_tests(test_results):
    tests = []
    for result in test_results:
        if result.start_time and result.end_time:
            duration = (result.end_time - result.start_time).total_seconds()
        else:
            duration = None
        tc = TestCase(
            name=result.name,
            elapsed_sec=duration,
            status=convert_status(result.status),
            timestamp=extract_timestamp(result))
        if result.exception:
            if TestStatus.failed == result.status:
                tc.add_failure_info(
                    message=exception_str(result.exception),
                    output=format_exception(result.exception))
            elif TestStatus.erred == result.status:
                tc.add_error_info(
                    message=exception_str(result.exception),
                    output=format_exception(result.exception))
            elif TestStatus.skipped == result.status:
                tc.add_skipped_info(
                    message=exception_str(result.exception))
        tests.append(tc)
    return tests


def ci_workspace_path():
    vars = dependency(os).environ
    config = config_for_module(__name__)
    key = config.ci_workspace_env_var
    if key in vars.keys():
        return vars[key]
    return None


def infer_package_name():
    """
    Use the path to the test script to infer a "package" name
    for the Junit report.
    """
    script = dependency(path_to_entry_script)()
    if not script:
        return ''
    script_path, _ = os.path.split(script)
    workspace_mask = ci_workspace_path()
    if workspace_mask:
        script_path = script_path.replace(workspace_mask, '')
    else:
        cwd_mask = dependency(os.getcwd)()
        script_path = script_path.replace(cwd_mask, '')
    name = script_path.replace('/', '.') + '.'
    if name.startswith('.'):
        name = name[1:]
    return name


class JunitReporter:

    REPORTS_DIRECTORY = 'reports'

    def __init__(self):
        self._dummy_test_case = None

    def activate(self):
        subscribe_event_handlers(self)

    def on_suite_erred(self, suite_name, exception=None, **kwargs):
        self._dummy_test_case = TestCase(name=suite_name, status='error')
        if exception:
            self._dummy_test_case.add_error_info(
                message=exception_str(exception),
                output=format_exception(exception))

    def on_suite_results_compiled(self, suite_results, **kwargs):
        suite_name = suite_results.suite_name or 'NamelessSuite'
        test_cases = convert_tests(suite_results.tests)
        if self._dummy_test_case:
            test_cases.append(self._dummy_test_case)
        suite = dependency(TestSuite)(
            name=infer_package_name() + suite_name,
            timestamp=current_time().isoformat(),
            test_cases=test_cases)
        xml_report = ElementTree.tostring(
            suite.build_xml_doc(), encoding='utf-8').decode(encoding='utf-8')
        EventBroker.publish(
            event=TestEvent.report_created,
            suite=suite,
            cases=test_cases,
            report_filename=suite_name + '.xml',
            report_content=xml_report)



from traceback import TracebackException


def indent(s):
    if s:
        spaces = 4 if s.startswith('File') or s.startswith('Traceback') else 8
        return ''.join([' ' for n in range(spaces)]) + s
    return ''


def format_exception(e):
    """
    Convert an exception to a human-readable message and stacktrace
    """
    tb = TracebackException.from_exception(e)
    stack = []
    for raw in tb.format():
        stack += [s.strip() for s in raw.split('\n')]
    msg = '(%s) %s' % ((type(e).__name__, e))
    trace = filter(lambda s: s, map(indent, [s for s in stack[:-1]]))
    return '{}\n'.format(msg) + '\n'.join(trace)



from ..common import precondition, skip  # noqa: F401
from ..common.activate_reporters import activate_reporters

from .do_and_check import do_and_check  # noqa: F401
from .test import test  # noqa: F401
from .test_suite import test_suite  # noqa: F401


activate_reporters()



def bind_attributes(obj, attrs):
    for k, v in attrs.items():
        if not isinstance(v, type) and '__call__' in dir(v):
            bind_method(obj, k, v)
        else:
            setattr(obj, k, v)


def bind_method(obj, name, func):
    setattr(obj, name, func.__get__(obj, obj.__class__))


class RunnableSuite:

    def __init__(self, attrs):
        self.__dict__['_suite_context'] = {}
        bind_attributes(self, attrs)

    def __getattr__(self, name):
        if name in self._suite_context.keys():
            return self._suite_context[name]
        if name in ['setup', 'setup_suite', 'teardown', 'teardown_suite']:
            # Default setup and teardown methods.
            # We can't define these with the typical "def" construct
            # because this would supercede __getattr__ and prevent
            # child classes from overriding these methods.
            return lambda: None
        raise AttributeError('I have no "%s"' % name)

    def __setattr__(self, name, value):
        self._suite_context[name] = value

    def get_suite_context(self):
        return self._suite_context

    def run_one_test(self, test_name):
        self.setup()
        try:
            getattr(self, test_name)()
        finally:
            self.teardown()



class AuthenticationFailed(RuntimeError):
    pass


class InvalidConfiguration(RuntimeError):
    pass


class InvalidHttpResponse(RuntimeError):
    pass


class NoSuchRecord(RuntimeError):
    pass


class TooManyRedirects(RuntimeError):
    """
    HttpClient followed a 3xx redirect which led to another, and another...
    """


class TestSkipped(Exception):
    pass


class TooManyThings(RuntimeError):
    """
    Something contains more entities than expected
    """


class UndefinedEvent(Exception):
    """
    Something referenced a test event not defined in TestEvent
    """



class TestResult:

    def __init__(self, test_name):
        self.artifacts = []
        self.end_time = None
        self.exception = None
        self.name = test_name
        self.start_time = None
        self.status = None


class SuiteResults:

    def __init__(self):
        self.artifacts = []
        self.suite_end_time = None
        self.suite_exception = None
        self.suite_name = None
        self.suite_start_time = None
        self.tests = []



import sys

from questions_three.vanilla import module_filename


def get_script():
    script_filename = module_filename(
        module=sys.modules[__name__], filename='dump_dom.js')
    with open(script_filename, 'r') as f:
        return f.read()


def dump_element(element):
    html = '<div>'
    html += '<b>%s</b> %s' % (element['name'], element['attributes'])
    if element['value']:
        html += '<p>%s</p>' % element['value']
    if element['children']:
        html += '<ul>'
        for child in element['children']:
            html += '<li>%s</li>' % dump_element(child)
        html += '</ul>'
    html += '</div>'
    return html


def dump_dom(driver):
    """
    Dump the DOM for the current page as an HTML string.

    driver -- (webdriver)
    """
    root = driver.execute_script(get_script())
    return '<html><body>%s</body></html>' % dump_element(root)



from datetime import datetime

from questions_three.constants import TestEvent, TestStatus
from questions_three.event_broker import EventBroker, \
    subscribe_event_handlers

from twin_sister import dependency

from .suite_results import SuiteResults, TestResult


def current_time():
    return dependency(datetime).now()


class ResultCompiler:

    def __init__(self):
        self.results = SuiteResults()

    def _find_or_create_test(self, test_name):
        for candidate in self.results.tests:
            if candidate.name == test_name:
                return candidate
        test = TestResult(test_name=test_name)
        test.start_time = current_time()
        test.status = TestStatus.running
        self.results.tests.append(test)
        return test

    def activate(self):
        subscribe_event_handlers(self)

    def on_artifact_created(self, artifact, test_name=None, **kwargs):
        if test_name:
            self._find_or_create_test(test_name).artifacts.append(artifact)
        else:
            self.results.artifacts.append(artifact)

    def on_suite_started(self, suite_name, **kwargs):
        self.results.suite_start_time = current_time()
        self.results.suite_name = suite_name

    def on_suite_ended(self, **kwargs):
        self.results.suite_end_time = current_time()
        EventBroker.publish(
            event=TestEvent.suite_results_compiled,
            suite_results=self.results)

    def on_suite_erred(self, exception=None, **kwargs):
        self.results.suite_exception = exception or Exception(
            'Suite erred.  No exception was provided.')

    def on_test_ended(self, test_name, **kwargs):
        test = self._find_or_create_test(test_name)
        test.end_time = current_time()
        if TestStatus.running == test.status:
            test.status = TestStatus.passed

    def on_test_erred(self, test_name, exception=None, **kwargs):
        test = self._find_or_create_test(test_name)
        test.status = TestStatus.erred
        test.exception = exception

    def on_test_failed(self, test_name, exception=None, **kwargs):
        test = self._find_or_create_test(test_name)
        test.status = TestStatus.failed
        test.exception = exception

    def on_test_skipped(self, test_name, exception=None, **kwargs):
        test = self._find_or_create_test(test_name)
        test.status = TestStatus.skipped
        test.exception = exception

    def on_test_started(self, test_name, **kwargs):
        self._find_or_create_test(test_name)



import sys

from twin_sister import dependency

from questions_three.module_cfg import config_for_module


def get_search_path():
    argv = dependency(sys.argv)
    if len(argv) > 1:
        return argv[1]
    return config_for_module(__name__).path_to_tests



class FormElementLacksAction(RuntimeError):
    """
    The form element lacks an "action" attribute
    """


class FormElementLacksMethod(RuntimeError):
    """
    The form element lacks a "method" attribute
    """


class FormNotFound(RuntimeError):
    """
    The specified HTML form was not found on the page
    """



from urllib.parse import urlencode, urljoin
from lxml import html


from .exceptions import FormNotFound, FormElementLacksAction, \
    FormElementLacksMethod


def assemble_submit_url(form_url, action):
    if action.startswith('http'):
        return action
    return urljoin(form_url, action)


class HtmlForm:

    def __init__(self, *, http_client, url, xpath):
        """
        Represents an HTML form on a Web page

        http_client -- HttpClient instance
        url -- GET this URL and expect it to contain an HTML form
        xpath -- Expect the form to be at this xpath
        """
        self.fields = {}
        self._http_client = http_client
        doc = html.fromstring(http_client.get(url).text)
        if xpath.startswith('/'):
            xpath = '.' + xpath
        forms = doc.findall(xpath)
        if not forms:
            raise FormNotFound('No form was found at the xpath')
        self._form = forms[0]
        if 'method' not in self._form.attrib.keys():
            raise FormElementLacksMethod('The form lacks a "method" attribute')
        self._method = self._form.attrib['method']
        if 'action' not in self._form.attrib.keys():
            raise FormElementLacksAction(
                'The form lacks an "action" attribute')
        self._submit_url = assemble_submit_url(
            url, self._form.attrib['action'])
        for element in self._form.findall('.//input'):
            keys = element.attrib.keys()
            if 'name' in keys and 'value' in keys:
                self.fields[element.attrib['name']] = element.attrib['value']

    def submit(self):
        func = getattr(self._http_client, self._method.lower())
        return func(
            self._submit_url,
            headers={'Content-type': 'application/x-www-form-urlencoded'},
            data=urlencode(self.fields))



#!/usr/bin/env python

import os
from setuptools import find_packages, setup
from setuptools.command.test import test
from unittest import TestLoader
from unittest.runner import TextTestRunner

MAJOR_VERSION = 2
# An even MINOR_VERSION number indicates a public release
MINOR_VERSION = 8
PATCH_VERSION = 4


class Tester(test):

    def run_tests(self):
        tests = TestLoader().discover('tests', pattern='test_*.py')
        runner = TextTestRunner()
        result = runner.run(tests)
        exit(0 if result.wasSuccessful() else 1)


def build_number():
    if 'BUILD_NUMBER' in os.environ.keys():
        return int(os.environ['BUILD_NUMBER'])
    return 0


with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    author='Mike Duskis',
    author_email='mike.duskis@cybergrx.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing'],
    cmdclass={'test': Tester},
    description='Toolkit for building automated integration checks',
    install_requires=[
        'boto3>=1.10.4',  # required to save artifacts to s3
	'docutils<0.15,>=0.10', #  part of the temporary work-around for botocore issue 1789
        'expects>=0.8.0',  # required by unit tests
        'junit-xml>=1.8',  # required by junit reporter
        'lxml>=4.1.1',  # required by html_form
        'PyYAML>=5.1',  # required by ModuleConfig
        'pyfakefs>=3.4.3',  # required by unit tests
        'requests>=2.18.4',  # required by HTTP Client
        'twin-sister>=4.2.6.0',  # required by unit tests
        'twine>=1.9.1',  # required by setup to upload package to Nexus
        'wheel>=0.30.0'  # required by setup to build package
        ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='questions-three',
    package_data={
        'questions_three': ['module_cfg.yml'],
        'questions_three.webdriver_tools.dom_dumper': ['dump_dom.js']
        },
    packages=find_packages(exclude=('tests.*', 'tests')),
    url='https://github.com/CyberGRX/questions-three',
    version='%d.%d.%d.%d' % (
        MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION, build_number()),
        )



from urllib.parse import urlparse

from questions_three.vanilla import url_append


def construct_redirect_url(*, request_url, response_location_header):
    parsed_request = urlparse(request_url)
    parsed_location = urlparse(response_location_header)
    if parsed_location.scheme:
        redirect = response_location_header
    else:
        redirect = url_append(
            f'{parsed_request.scheme}://{parsed_request.netloc}',
            response_location_header)
    parsed_redirect = urlparse(redirect)
    if parsed_request.fragment and not parsed_redirect.fragment:
        redirect += f'#{parsed_request.fragment}'
    return redirect



from questions_three.exceptions import InvalidConfiguration


def numeric_value(*, key, val):
    try:
        return float(val)
    except ValueError as e:
        raise InvalidConfiguration(
            '"%s" was configured with a numeric value.  '
            'Refusing to override it with non-number "%s"'
            % (key, val)) from e


def bool_value(*, key, val):
    lc_val = str(val).lower()
    if lc_val in ('true', 'y', 'yes', '1'):
        return True
    if lc_val in ('false', 'n', 'no', '0'):
        return False
    raise InvalidConfiguration(
        '"%s" was configured with a boolean value.  '
        'Refusing to override it with "%s"'
        % (key, val))


def transform_type(*, key, val, new_type):
    if new_type == bool:
        return bool_value(key=key, val=val)
    if new_type in (int, float):
        return numeric_value(key=key, val=val)
    return val



from twin_sister import dependency

from questions_three.event_broker import subscribe_event_handlers
from questions_three.logging import logger_for_module
from questions_three.vanilla import format_exception


class EventLogger:

    def __init__(self):
        self._log = dependency(logger_for_module)(__name__)

    def activate(self):
        subscribe_event_handlers(self)

    def on_suite_ended(self, suite_name, **kwargs):
        self._log.info('Suite "%s" ended' % suite_name)

    def on_suite_erred(self, exception=None, **kwargs):
        self._log.error(format_exception(exception))

    def on_suite_started(self, suite_name, **kwargs):
        self._log.info('Suite "%s" started' % suite_name)

    def on_test_ended(self, test_name, **kwargs):
        self._log.info('Test "%s" ended' % test_name)

    def on_test_erred(self, exception, **kwargs):
        self._log.error(format_exception(exception))

    def on_test_failed(self, test_name, exception, **kwargs):
        self._log.warning('Test "%s" failed: %s' % (test_name, exception))

    def on_test_skipped(self, test_name, exception, **kwargs):
        self._log.warning('Test "%s" skipped: %s' % (test_name, exception))

    def on_test_started(self, test_name, **kwargs):
        self._log.info('Test "%s" started' % test_name)



class HttpError(RuntimeError):
    status_group = None
    status_code = None

    def __init__(self, *args, response=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.response = response


class HttpRedirect(HttpError):
    status_group = 3


class HttpMultipleChoices(HttpRedirect):
    status_code = 300


class HttpMovedPermanently(HttpRedirect):
    status_code = 301


class HttpFound(HttpRedirect):
    status_code = 302


class HttpSeeOther(HttpRedirect):
    status_code = 303


class HttpUseProxy(HttpRedirect):
    status_code = 305


class HttpTemporaryRedirect(HttpRedirect):
    status_code = 307


# https://tools.ietf.org/html/rfc7238
class PermanentRedirect(HttpRedirect):
    status_code = 308


class HttpClientError(HttpError):
    status_group = 4


class HttpBadRequest(HttpClientError):
    status_code = 400


class HttpUnauthorized(HttpClientError):
    status_code = 401


class HttpPaymentRequired(HttpClientError):
    status_code = 402


class HttpForbidden(HttpClientError):
    status_code = 403


class HttpNotFound(HttpClientError):
    status_code = 404


class HttpMethodNotAllowed(HttpClientError):
    status_code = 405


class HttpNotAcceptable(HttpClientError):
    status_code = 406


class HttpProxyAuthenticationRequired(HttpClientError):
    status_code = 407


class HttpRequestTimeout(HttpClientError):
    status_code = 408


class HttpConflict(HttpClientError):
    status_code = 409


class HttpGone(HttpClientError):
    status_code = 410


class HttpLengthRequired(HttpClientError):
    status_code = 411


class HttpPayloadTooLarge(HttpClientError):
    status_code = 413


class HttpUriTooLong(HttpClientError):
    status_code = 414


class HttpUnsupportedMediaType(HttpClientError):
    status_code = 415


class HttpExpectationFailed(HttpClientError):
    status_code = 417


class HttpImATeapot(HttpClientError):
    status_code = 418


class HttpUpgradeRequired(HttpClientError):
    status_code = 426


class HttpServerError(HttpError):
    status_group = 5


class HttpInternalServerError(HttpServerError):
    status_code = 500


class HttpNotImplemented(HttpServerError):
    status_code = 501


class HttpBadGateway(HttpServerError):
    status_code = 502


class HttpServiceUnavailable(HttpServerError):
    status_code = 503


class HttpGatewayTimeout(HttpServerError):
    status_code = 504


class HttpHttpVersionNotSupported(HttpServerError):
    status_code = 505



from datetime import datetime, timedelta
import os
from time import sleep

from twin_sister import dependency

from questions_three.logging import logger_for_module
from questions_three.module_cfg import config_for_module

from .pool import Pool


log = logger_for_module(__name__)


def discover(top_dir):
    walk = dependency(os).walk
    scripts = []
    for path, _, files in walk(top_dir):
        scripts += [
            dependency(os).path.join(path, fn)
            for fn in files
            if not fn.startswith('_') and fn.endswith('.py')]
    return scripts


def _float_or_none(name, value):
    if value in (None, ''):
        return None
    try:
        return float(value)
    except ValueError as e:
        raise TypeError(
            f'Expected {name} ({value}) to be a number') from e


def run_all(path):
    """
    Run all tests beneath the given path
    """
    cfg = config_for_module(__name__)
    throttle = cfg.delay_between_checks_for_parallel_suite_completion
    limit = cfg.max_parallel_suites
    timeout = _float_or_none(
            name='$RUN_ALL_TIMEOUT', value=cfg.run_all_timeout)
    if timeout is None:
        expiry = None
    else:
        expiry = dependency(datetime).now() + timedelta(
            seconds=_float_or_none(
                name='$RUN_ALL_TIMEOUT', value=cfg.run_all_timeout))
    pool = Pool(limit)
    queue = list(discover(path))
    while queue:
        if expiry and dependency(datetime).now() > expiry:
            break
        while pool.has_capacity() and queue:
            pool.add(queue.pop(0))
        log.debug('Waiting in queue: %d' % len(queue))
        sleep(throttle)
    pool.wait_for_jobs(expiry=expiry)
    return pool.failure_count



from .meta_suite import MetaSuite


class TestSuite(metaclass=MetaSuite):
    """
    The exposed part of a self-running xUnit test suite

    Test suites declare TestSuite as their parent class
    """
